/*
 * avi.h library for AVI file format i/o
 * Copyright (C) 2000 Arne Schirmacher <arne@schirmacher.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/** Common AVI declarations
 
    Some of this comes from the public domain AVI specification, which
    explains the microsoft-style definitions.
 
    \file avi.h
*/
/** abc
    \page avipage The AVI file format: theory of operation
 
*/


#ifndef _AVI_H
#define _AVI_H 1

#include "frame.h"

#define AVI_INDEX_OF_INDEXES (0x00)
#define AVI_INDEX_OF_CHUNKS (0x01)
#define AVI_INDEX_2FIELD (0x01)

enum { AVI_PAL, AVI_NTSC, AVI_AUDIO_48KHZ, AVI_AUDIO_44KHZ, AVI_AUDIO_32KHZ };

/** Declarations of the main AVI file header
 
    The contents of this struct goes into the 'avih' chunk.  */

typedef struct
{
    /// frame display rate (or 0L)
    DWORD dwMicroSecPerFrame;
    /// max. transfer rate
    DWORD dwMaxBytesPerSec;
    /// pad to multiples of this size, normally 2K
    DWORD dwPaddingGranularity;
    /// the ever-present flags
    DWORD dwFlags;
    /// # frames in file
    DWORD dwTotalFrames;
    DWORD dwInitialFrames;
    DWORD dwStreams;
    DWORD dwSuggestedBufferSize;

    DWORD dwWidth;
    DWORD dwHeight;

    DWORD dwReserved[4];
}
MainAVIHeader;

typedef struct
{
    WORD top, bottom, left, right;
}
RECT;

/** Declaration of a stream header
 
    The contents of this struct goes into the 'strh' header. */

typedef struct
{
    FOURCC fccType;
    FOURCC fccHandler;
    DWORD dwFlags;                /* Contains AVITF_* flags */
    WORD wPriority;
    WORD wLanguage;
    DWORD dwInitialFrames;
    DWORD dwScale;
    DWORD dwRate;                 /* dwRate / dwScale == samples/second */
    DWORD dwStart;
    DWORD dwLength;               /* In units above... */
    DWORD dwSuggestedBufferSize;
    DWORD dwQuality;
    DWORD dwSampleSize;
    RECT rcFrame;
}
AVIStreamHeader;

typedef struct
{
    DWORD dwDVAAuxSrc;
    DWORD dwDVAAuxCtl;
    DWORD dwDVAAuxSrc1;
    DWORD dwDVAAuxCtl1;
    DWORD dwDVVAuxSrc;
    DWORD dwDVVAuxCtl;
    DWORD dwDVReserved[2];
}
DVINFO;

typedef struct
{
    DWORD dwSize;
    LONG biWidth;
    LONG biHeight;
    WORD biPlanes;
    WORD biBitCount;
    DWORD biCompression;
    DWORD biSizeImage;
    LONG biXPelsPerMeter;
    LONG biYPelsPerMeter;
    DWORD biClrUsed;
    DWORD biClrImportant;
}
BITMAPINFOHEADER;

typedef struct
{
    WORD wFormatTag;
    WORD nChannels;
    DWORD nSamplesPerSec;
    DWORD nAvgBytesPerSec;
    WORD nBlockAlign;
    WORD wBitsPerSample;
    //WORD  cbSize;
}




WAVEFORMATEX;

typedef struct
{
    WORD wLongsPerEntry;
    BYTE bIndexSubType;
    BYTE bIndexType;
    DWORD nEntriesInUse;
    DWORD dwChunkId;
    DWORD dwReserved[3];
    struct avisuperindex_entry
    {
        QUADWORD qwOffset;
        DWORD dwSize;
        DWORD dwDuration;
    }
    aIndex[2014];
}
AVISuperIndex;

typedef struct
{
    WORD wLongsPerEntry;
    BYTE bIndexSubType;
    BYTE bIndexType;
    DWORD nEntriesInUse;
    DWORD dwChunkId;
    QUADWORD qwBaseOffset;
    DWORD dwReserved;
    struct avifieldindex_entry
    {
        DWORD dwOffset;
        DWORD dwSize;
    }
    aIndex[4028];
}
AVIStdIndex;

typedef struct
{
    struct avisimpleindex_entry
    {
        FOURCC	dwChunkId;
        DWORD	dwFlags;
        DWORD	dwOffset;
        DWORD	dwSize;
    }
    aIndex[4000];
    DWORD	nEntriesInUse;
}
AVISimpleIndex;

typedef struct
{
    DWORD dirEntryType;
    DWORD dirEntryName;
    DWORD dirEntryLength;
    size_t dirEntryOffset;
    int dirEntryWrittenFlag;
    int dirEntryParentList;
}
AviDirEntry;


/** base class for all AVI type files
 
    It contains methods and members which are the same in all AVI type files regardless of the particular compression, number
    of streams etc. 
 
    The AVIFile class also contains methods for handling several indexes to the video frame content. */

class AVIFile : public RIFFFile
{
protected:
    MainAVIHeader mainHdr;
    AVISuperIndex indx;
    AVIStdIndex ix00;
    AVISimpleIndex idx1;

    int	index_type;
    int current_ix00;
    int file_list;
    int riff_list;
    int hdrl_list;
    int avih_chunk;
    int movi_list;
    int indx_chunk;
    int ix00_chunk;
    int idx1_chunk;
    int junk_chunk;

public:
    AVIFile();
    AVIFile(const char *s);
    virtual ~AVIFile();

    virtual void Init(int format, int sampleFrequency);
    virtual int GetDVFrameInfo(int &offset, int &size, int frameNum);
    virtual int GetDVFrame(Frame &frame, int frameNum);
    virtual int GetTotalFrames() const;
    virtual void PrintDirectoryEntryData(const RIFFDirEntry &entry) const;
    virtual void WriteFrame(const Frame &frame);
    virtual void ParseList(int parent);
    virtual void ParseRIFF(void);
    virtual void ReadIndex(void);
    virtual void WriteRIFF(void);
    virtual void FlushIX00Index(void);
    virtual void UpdateIndex(int frame_chunk);
};


/** writing Type 1 DV AVIs
 
*/

class AVI1File : public AVIFile
{
private:
    AVIStreamHeader stream1Hdr;
    DVINFO dvinfo;
    int dmlh[62];

    int strl1_list;
    int strh1_chunk;
    int strf1_chunk;
    int odml_list;
    int dmlh_chunk;

public:
    AVI1File();
    AVI1File(const char *s);
    virtual ~AVI1File();

    virtual void Init(int format, int sampleFrequency);
    virtual void WriteFrame(const Frame &frame);
    virtual void WriteRIFF(void);
};


/** writing Type 2 (separate audio data) DV AVIs
 
This file type contains both audio and video tracks. It is therefore more compatible
to certain Windows programs, which expect any AVI having both audio and video tracks.
The video tracks contain the raw DV data (as in type 1) and the extracted audio tracks.
 
Note that because the DV data contains audio information anyway, this means duplication
of data and a slight increase of file size.
 
*/

class AVI2File : public AVIFile
{
private:
    AVIStreamHeader stream1Hdr;
    BITMAPINFOHEADER bitmapinfo;
    AVIStreamHeader stream2Hdr;
    WAVEFORMATEX waveformatex;

    int strl1_list;
    int strh1_chunk;
    int strf1_chunk;
    int strl2_list;
    int strh2_chunk;
    int strf2_chunk;
public:
    AVI2File();
    AVI2File(const char *s);
    virtual ~AVI2File();

    virtual void Init(int format, int sampleFrequency);
    virtual void WriteFrame(const Frame &frame);
    virtual void WriteRIFF(void);
};
#endif


