/*
 * riff.cc library for RIFF file format i/o
 * Copyright (C) 2000 Arne Schirmacher <arne@schirmacher.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "riff.h"
#include <fcntl.h>
#include <unistd.h>

#include <string>
#include <iostream>
#include <strstream>
#include <iomanip>

/** \defgroup utils Utilities
  
    A few small utilities which do not fit in any class. */


/** \defgroup dir Directory
  
    These methods are for adding, modifying and retrieving entries
    from the directory.
 
    \todo We should also have a delete entry method, but probably nobody needs it.  */


/** \defgroup status Status
 
    Methods for getting some information about the directory entry. */


/** \defgroup read Reading a RIFF file
 
    Parsing and reading the directory structure of a  RIFF file. */


/** \defgroup write Writing a RIFF file
 
    Writing the directory structure to a RIFF file. */


/** error handler for negative result codes
 
    Whenever this is called with a negative argument, it will throw an
    exception. Typically used with operating system functions.
 
    \ingroup utils
*/

void real_fail_neg(int i, char *file, int line)
{
    if (i < 0) {

    	string exc;
	strstream sb;
	
	sb << "exception in " << file << ", line " << line << ": function returned " << i;
	sb >> exc;
        throw exc;
    }
}


/** error handler for NULL result codes
 
    Whenever this is called with a NULL argument, it will throw an
    exception. Typically used with functions like malloc() and new().
 
   \ingroup utils 
*/

void real_fail_null(void *p, char *file, int line)
{
    if (p == NULL) {

    	string exc;
	strstream sb;

	sb << "exception in " << file << ", line " << line << ": function returned NULL";
	sb >> exc;
        throw exc;
    }
}


/** make a 32 bit "string-id"
 
    \ingroup utils 
    \param s a pointer to 4 chars
    \return the 32 bit "string id"
    \bugs This code is byte-order dependent and runs only on little-endian (Intel) systems
    \bugs It is not checked whether we really have 4 characters
 
    Some compilers understand constants like int id = 'ABCD'; but I
    could not get it working on the gcc compiler so I had to use this
    workaround. We can now use id = make_fourcc("ABCD") instead. */

FOURCC make_fourcc(char *s)
{
    if (s[0] == 0)
        return 0;
    else
        return (s[3] << 24) | (s[2] << 16) | (s[1] << 8) | s[0];
}


RIFFDirEntry::RIFFDirEntry()
{}


RIFFDirEntry::RIFFDirEntry(FOURCC t, FOURCC n, int l, int o, int p): type(t), name(n), length(l), offset(o), parent(p), written(0)
{}


/** Creates the object without an output file.
 
    \todo This is useless. Delete it.
*/

RIFFFile::RIFFFile()
{
    fd = -1;
}


/** Creates the object with its output file.
 
    \param s the filename
    \note We need also a way to open in read only and write only mode.
    \bug this should really go into an Init function. The open call
    might fail. If it fails, it will assign -1 to fd, which prevents
    the destructor from trying to close the file. */

RIFFFile::RIFFFile(const char *s)
{
    fail_neg(fd = open(s, O_RDWR | O_NONBLOCK | O_CREAT, 00644));
}


/** Destroys the object.
 
    If it has an associated opened file, close it. */

RIFFFile::~RIFFFile()
{
    if (fd != -1)
        close(fd);
}


/** Creates the object with its output file.
 
    \param s the filename
    \note We need also a way to open in read only and write only mode.
    \bug this should really go into an Init function. The open call
    might fail. If it fails, it will assign -1 to fd, which prevents
    the destructor from trying to close the file. */

void RIFFFile::Open(char *s)
{
    fail_neg(fd = open(s, O_RDWR | O_NONBLOCK | O_CREAT, 00644));
}


/** Destroys the object.
 
    If it has an associated opened file, close it. */

void RIFFFile::Close()
{
    if (fd != -1) {
        close(fd);
        fd = -1;
    }
}

/** Adds an entry to the list of containers.
 
    \ingroup dir
    \param type the type of this entry
    \param name the name
    \param length the length of the data in the container
    \param list the container in which this object is contained. 
    \return the ID of the newly created entry
 
    The topmost object is not contained in any other container. Use
    the special ID RIFF_NO_PARENT to create the topmost object. */

int RIFFFile::AddDirectoryEntry(FOURCC type, FOURCC name, int length, int list)
{
    /* Put all parameters in an RIFFDirEntry object. The offset is
       currently unknown. */

    RIFFDirEntry entry(type, name, length, 0 /* offset */, list);

    /* If the new chunk is in a list, then get the offset and size
       of that list. The offset of this chunk is the end of the list
       (parent_offset + parent_length) plus the size of the chunk
       header. */

    if (list != RIFF_NO_PARENT) {
        RIFFDirEntry parent = GetDirectoryEntry(list);
        entry.offset = parent.offset + parent.length + RIFF_HEADERSIZE;
    }

    /* The list which this new chunk is a member of has now increased in
       size. Get that directory entry and bump up its length by the size
       of the chunk. Since that list may also be contained in another
       list, walk up to the top of the tree. */

    while (list != RIFF_NO_PARENT) {
        RIFFDirEntry parent = GetDirectoryEntry(list);
        parent.length += RIFF_HEADERSIZE + length;
        SetDirectoryEntry(list, parent);
        list = parent.parent;
    }

    directory.insert(directory.end(), entry);

    return directory.size() - 1;
}


/** Modifies an entry.
 
    \ingroup dir
    \param i the ID of the entry which is to modify
    \param type the type of this entry
    \param name the name
    \param length the length of the data in the container
    \param list the container in which this object is contained.
    \note Do not change length, offset, or the parent container.
    \note Do not change an empty name ("") to a name and vice versa */

void RIFFFile::SetDirectoryEntry(int i, FOURCC type, FOURCC name, int length, int offset, int list)
{
    RIFFDirEntry entry(type, name, length, offset, list);

    assert(i >= 0 && i < (int)directory.size());

    directory[i] = entry;
}


/** Modifies an entry.
 
    The entry.written flag is set to false because the contents has been modified
 
    \ingroup dir
    \param i the ID of the entry which is to modify
    \param entry the new entry 
    \note Do not change length, offset, or the parent container.
    \note Do not change an empty name ("") to a name and vice versa */

void RIFFFile::SetDirectoryEntry(int i, RIFFDirEntry &entry)
{
    assert(i >= 0 && i < (int)directory.size());

    entry.written = false;
    directory[i] = entry;
}


/** Retrieves an entry.
 
    Gets the most important member variables.
 
    \ingroup dir
    \param i the ID of the entry to retrieve
    \param type
    \param name
    \param length
    \param offset
    \param list */

void RIFFFile::GetDirectoryEntry(int i, FOURCC &type, FOURCC &name, int &length, int &offset, int &list) const
{
    RIFFDirEntry entry;

    assert(i >= 0 && i < (int)directory.size());

    entry = directory[i];
    type = entry.type;
    name = entry.name;
    length = entry.length;
    offset = entry.offset;
    list = entry.parent;
}


/** Retrieves an entry.
 
    Gets the whole RIFFDirEntry object.
 
    \ingroup dir
    \param i the ID of the entry to retrieve
    \return the entry */

RIFFDirEntry RIFFFile::GetDirectoryEntry(int i) const
{
    assert(i >= 0 && i < (int)directory.size());

    return directory[i];
}


/** Calculates the total size of the file
 
    \ingroup status
    \return the size the file in bytes
*/

int RIFFFile::GetFileSize(void) const
{

    /* If we have at least one entry, return the length field
       of the FILE entry, which is the length of its contents,
       which is the actual size of whatever is currently in the
       AVI directory structure. 

       Note that the first entry does not belong to the AVI
       file.

       If we don't have any entry, the file size is zero. */

    if (directory.size() > 0)
        return directory[0].length;
    else
        return 0;
}


/** prints the attributes of the entry
 
    \ingroup status
    \param i the ID of the entry to print
*/

void RIFFFile::PrintDirectoryEntry(int i) const
{
    RIFFDirEntry    entry;
    RIFFDirEntry    parent;
    FOURCC          entry_name;
    FOURCC          list_name;
		    
    /* Get all attributes of the chunk object. If it is contained
       in a list, get the name of the list too (otherwise the name of
       the list is blank). If the chunk object doesnt have a name (only
       LISTs and RIFFs have a name), the name is blank. */

    entry = GetDirectoryEntry(i);
    if (entry.parent != RIFF_NO_PARENT) {
        parent = GetDirectoryEntry(entry.parent);
        list_name = parent.name;
    } else {
        list_name = make_fourcc("    ");
    }
    if (entry.name != 0) {
        entry_name = entry.name;
    } else {
        entry_name = make_fourcc("    ");
    }

    /* Print out the ascii representation of type and name, as well as
       length and file offset. */

    cout << hex << setfill('0') << "type: "
    << (char)((entry.type >> 0) & 0x000000ff)
    << (char)((entry.type >> 8) & 0x000000ff)
    << (char)((entry.type >> 16) & 0x000000ff)
    << (char)((entry.type >> 24) & 0x000000ff)
    << " name: "
    << (char)((entry_name >> 0) & 0x000000ff)
    << (char)((entry_name >> 8) & 0x000000ff)
    << (char)((entry_name >> 16) & 0x000000ff)
    << (char)((entry_name >> 24) & 0x000000ff)
    << " length: 0x" << setw(8) << entry.length 
    << " offset: 0x" << setw(8) << (int)entry.offset 
    << " list: "
    << (char)((list_name >> 0) & 0x000000ff)
    << (char)((list_name >> 8) & 0x000000ff)
    << (char)((list_name >> 16) & 0x000000ff)
    << (char)((list_name >> 24) & 0x000000ff) << dec << endl;

    /* print the content itself */

    PrintDirectoryEntryData(entry);
}


/** prints the contents of the entry
 
    Prints a readable representation of the contents of an index.
    Override this to print out any objects you store in the RIFF file.
 
    \ingroup status
    \param entry the entry to print */

void RIFFFile::PrintDirectoryEntryData(const RIFFDirEntry &entry) const
    {}


/** prints the contents of the whole directory
 
    Prints a readable representation of the contents of an index.
    Override this to print out any objects you store in the RIFF file.
 
    \ingroup status
    \param entry the entry to print */

void RIFFFile::PrintDirectory() const
{
    int i;
    int count = directory.size();

    for (i = 0; i < count; ++i)
        PrintDirectoryEntry(i);
}


/** finds the index
 
    finds the index of a given directory entry type 
 
    \ingroup dir
    \todo we should also have an index to find the nth entry of a type
    \todo inefficient if the directory has lots of items
    \param type the type of the entry to find
    \return the index of the found object in the directory, or -1 if not found */

int RIFFFile::FindDirectoryEntry(FOURCC type) const
{
    int i;
    int count = directory.size();

    for (i = 0; i < count; ++i)
        if (directory[i].type == type)
            return i;

    return -1;
}


/** Reads all items that are contained in one list
 
    Read in one chunk and add it to the directory. If the chunk
    happens to be of type LIST, then call ParseList recursively for
    it.
 
    \param parent The id of the item to process
    \ingroup read 
*/

void RIFFFile::ParseChunk(int parent)
{
    FOURCC type;
    int length;
    int pos;

    /* Check whether it is a LIST. If so, let ParseList deal with it */

    read(fd, &type, sizeof(type));
    if (type == make_fourcc("LIST")) {
        fail_neg(lseek(fd, -sizeof(type), SEEK_CUR));
        ParseList(parent);
    }

    /* it is a normal chunk, create a new directory entry for it */

    else {
        fail_neg(read(fd, &length, sizeof(length)));
        if (length & 1) length++;
        fail_neg(pos = lseek(fd, 0, SEEK_CUR));
        AddDirectoryEntry(type, 0, length, parent);
        fail_neg(lseek(fd, length, SEEK_CUR));
    }
}


/** Reads all items that are contained in one list
 
    \param parent The id of the list to process
 
    \ingroup read
*/

void RIFFFile::ParseList(int parent)
{
    FOURCC type;
    FOURCC name;
    int length;
    int pos;
    int list;
    int listEnd;

    /* Read in the chunk header (type and length). */

    fail_neg(read(fd, &type, sizeof(type)));
    fail_neg(read(fd, &length, sizeof(length)));
    if (length & 1) length++;

    /* The contents of the list starts here. Obtain its offset. The list
       name (4 bytes) is already part of the contents). */

    fail_neg(pos = lseek(fd, 0, SEEK_CUR));
    fail_neg(read(fd, &name, sizeof(name)));

    /* Add an entry for this list. */

    list = AddDirectoryEntry(type, name, sizeof(name), parent);

    /* Read in any chunks contained in this list. This list is the
       parent for all chunks it contains. */

    listEnd = pos + length;
    while (pos < listEnd) {
        ParseChunk(list);
        fail_neg(pos = lseek(fd, 0, SEEK_CUR));
    }
}


/** Reads the directory structure of the whole RIFF file
 
    \ingroup read
*/

void RIFFFile::ParseRIFF(void)
{
    int container = AddDirectoryEntry(make_fourcc("FILE"), make_fourcc("FILE"), 0, RIFF_NO_PARENT);

    off_t filesize = lseek(fd, 0, SEEK_END);
    lseek(fd, 0, SEEK_SET);

    /* There should be at least one list ('AVI ') here, sometimes followed by 'AVIX' lists */

    while (lseek(fd, 0, SEEK_CUR) < filesize) {
        ParseList(container);
    }
}


/** Reads one item including its contents from the RIFF file
 
    \param chunk_index The index of the item to write
    \param data A pointer to the data
 
    \ingroup read
*/

void RIFFFile::ReadChunk(int chunk_index, void *data)
{
    RIFFDirEntry entry;

    entry = GetDirectoryEntry(chunk_index);
    fail_neg(lseek(fd, entry.offset, SEEK_SET));
    fail_neg(read(fd, data, entry.length));
}


/** Writes one item including its contents to the RIFF file
 
    \param chunk_index The index of the item to write
    \param data A pointer to the data
 
    \ingroup write
*/

void RIFFFile::WriteChunk(int chunk_index, const void *data)
{
    RIFFDirEntry entry;

    entry = GetDirectoryEntry(chunk_index);
    fail_neg(lseek(fd, entry.offset - RIFF_HEADERSIZE, SEEK_SET));
    fail_neg(write(fd, &entry.type, sizeof(entry.type)));
    fail_neg(write(fd, &entry.length, sizeof(entry.length)));
    fail_neg(write(fd, data, entry.length));

    /* Remember that this entry already has been written. */

    directory[chunk_index].written = true;
}


/** Writes out the directory structure
 
    For all items in the directory list that have not been written
    yet, it seeks to the file position where that item should be
    stored and writes the type and length field. If the item has a
    name, it will also write the name field.
 
    \ingroup write
    \note It does not write the contents of any item. Use WriteChunk to do that. */

void RIFFFile::WriteRIFF(void)
{
    int i;
    RIFFDirEntry entry;
    int count = directory.size();

    /* Start at the second entry (RIFF), since the first entry (FILE)
       is needed only for internal purposes and is not written to the
       file. */

    for (i = 1; i < count; ++i) {

        /* Only deal with entries that havent been written */

        entry = GetDirectoryEntry(i);
        if (entry.written == false) {

            /* A chunk entry consist of its type and length, a list
               entry has an additional name. Look up the entry, seek
               to the start of the header, which is at the offset of
               the data start minus the header size and write out the
               items. */

            fail_neg(lseek(fd, entry.offset - RIFF_HEADERSIZE, SEEK_SET));
            fail_neg(write(fd, &entry.type, sizeof(entry.type)));
            fail_neg(write(fd, &entry.length, sizeof(entry.length)));

            /* If it has a name, it is a list. Write out the extra name
               field. */

            if (entry.name != 0) {
                fail_neg(write(fd, &entry.name, sizeof(entry.name)));
            }

            /* Remember that this entry already has been written. */

            directory[i].written = true;
        }
    }
}

